#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include "firefly.h"

const prog_uint8_t about[] = "Jar of Fireflies\n"
	    "Design and Implimentation by Xander Hudson (jar@synoptic.org)\n"
	    "Idea and Inspiration by Kayobi Tierney\n"
	    "Special thanks to Katie Horn for getting me to think about electronics";
const prog_uint8_t version[] = "$Revision: 1.41 $ $Date: 2007/01/10 04:35:55 $";

#define LEDS_OFF 0

#define WATCHDOG_OK  WDTCR |= _BV(WDIE)

const Song *pickasong (void);
uint8_t randwaitval (void);
uint8_t randbits (uint8_t);
uint8_t randscaleval (void);
void showbootup (void);

#define ch1playing ((volatile io_reg*) _SFR_MEM_ADDR(FLAGS0))->b0
#define ch1resting ((volatile io_reg*) _SFR_MEM_ADDR(FLAGS0))->b1
#define need_mood ((volatile io_reg*) _SFR_MEM_ADDR(FLAGS0))->b4
#define ch1noteready ((volatile io_reg*) _SFR_MEM_ADDR(FLAGS0))->b6

volatile uint8_t ch1restcount;

volatile uint8_t ch1nextnote;

volatile uint8_t ch1scale;
volatile uint8_t ch1nextscale;

register uint8_t ch1bright asm ("r2");
register uint8_t ch1err asm ("r4");
register uint8_t ch1pin asm ("r6");
register uint8_t portval asm ("r8");

/*  mood, mood_count						    */
/*  This value changes slowly over time and is added to the random	    */
/*  wait timer.  Having this change slowly adds a sense of 'moodiness'	    */
/*  where the fireflys will go through phases of being active or lethargic. */

uint8_t mood = 2;
uint16_t mood_count = 30;   // WDT cycles before changing mood

/* Intentionally crash the program leaving a visual clue.  The	*/
/* watchdog should then come and reboot us.			*/
#define FREAKOUT  for (;;) {PORTB = LEDS_OFF; PORTB = ~LEDS_OFF; } 

ISR(WDT_vect) {
    if (ch1resting && (--ch1restcount == 0)) ch1resting = 0;
    if (!need_mood && (--mood_count == 0)) need_mood = 1;
}

ISR(TIM0_OVF_vect) {
    static uint8_t x = 25;

    if ( x-- == 0 ) {
	x = 25;
	if (ch1playing) {
	    // We're playing a song. Is there another note
	    // for us to read? If not, shut down.
	    if (ch1noteready) {  
		ch1bright = ch1nextnote;
		ch1noteready = 0;
	    } else {
		// We ran out of notes to play.
		ch1bright = 0;
		// Has our rest duration been chosen yet?
		if (ch1restcount != 0) {  
		    ch1playing = 0;   
		    ch1resting = 1;
		} // if not, go through the loop again.
	    }
	} else if (!ch1resting && ch1noteready) {  
		// We weren't playing a song but now we see
		// that we're clear to start.
		ch1bright = ch1nextnote;
		ch1noteready = 0;
		ch1playing = 1;
	}

    }

    if (ch1playing) OCR0A = ~ch1bright;

}


/*
 * Do all the startup-time peripheral initializations.
 */
static void
ioinit(void)
{ 
    wdt_enable(WDTO_500MS);   // Set watchdog timeout

    WDTCR |= _BV(WDIE);    // Set watchdog to generate interrupt.

    // Set all pins as outputs.
    // Drive PIN_A and PIN_B low, everybody else high.
    DDRB = _BV(0) | _BV(1);
    PORTB= LEDS_OFF;

    PRR |= _BV(PRTIM1);  // Turn off Timer1 to save power.
    PRR |= _BV(PRUSI);	 // Turn off USI clock
    PRR |= _BV(PRADC);	 // Turn off ADC

    /*
     * Set up the 8-bit timer 0.
     *
     * Timer0 will be set in CTC up with a prescalar of 8 and an OCR0A copmare 
     * value of 25. 
     * A core clock frequency of 8mhz / 4 means we should get an interrupt 
     * frequency of about 39hz.  (freq = 976.563 / (OCR0A))
     */

    TCCR0A = _BV(COM0A1) | _BV(COM0A0) | _BV(WGM00) | _BV(WGM01);  // Fast PWM mode, inverting, ocr0a

    //TCCR0B = _BV(CS01) | _BV(CS00); // Clock/64
    //TCCR0B = _BV(CS00); // Clock
    TCCR0B = _BV(CS01); // Clock/8
    //OCR0A = 25;

    clock_prescale_set(clock_div_4);

    TIMSK = _BV(TOIE0);

    OCR0A = 0xff;
    WATCHDOG_OK;
}

int main(void) {
    const Song *ch1song = 0;
    uint16_t ch1noteptr = 0;
    uint8_t temp;

    ioinit();

    // Zero out register variables.  ch1bright = ch1err = 
	ch1pin = portval = 0;

    // Zero out flags.
    FLAGS0 = 0;

    sei();

    showbootup();

    ch1restcount = randwaitval();

    ch1resting = 1;

    for (;;) {

	/* If we have a channel playing, take care of its    */
	/* housekeeping first.				     */
	if (ch1playing) {
	    /* If channel is playing but doesn't have its next note   */
	    /* loaded, load it.					      */

	    if (ch1playing && !ch1noteready) {
		if ( ch1noteptr != ch1song->notecount ) {
		
		    temp = pgm_read_byte(&ch1song->notes[ch1noteptr++]);
		    temp = ((uint16_t)(ch1scale * temp)) >> 8;

		    cli();
		    ch1nextnote = temp;
		    ch1noteready = 1;
		    sei();

		    continue;  // Loop over from start.
		} // Else end of song.
	    }

	    /* If a channel is playing a song and its restcount is 0,	*/
	    /* we need to pick a new random restcount for it.		*/

	    if (ch1playing && ch1restcount == 0) {
		temp = randwaitval();
		cli(); 
		ch1restcount = temp; 
		sei();
		continue;  // Loop over from start.
	    }
	}

	/* If channel is done resting and waiting for a song to   */
	/* be queued up, select a song for it.			  */
	if (!ch1playing && !ch1resting && !ch1noteready) {   
	    ch1song = pickasong();	     
	    ch1noteptr = 0;

	    // Now pick a random scaling value so the brightness
	    // varies over time.
	    ch1scale = randscaleval();
	   
	    temp = pgm_read_byte(&ch1song->notes[ch1noteptr++]);
	    temp = ((uint16_t)(ch1scale * temp)) >> 8;

	    // If the other channel isn't playing lets reselect the
	    // master pin.

	    cli();
	    ch1nextnote = temp;
	    ch1noteready = 1;
	    sei();

	    continue;  // Loop over from start.
	}

	if (need_mood) {
#ifdef PHOTOSHOOT
	    mood = 1; 
#else			
	    mood = randbits(5) + 8; 
#endif
	    mood_count = 60 * 5 * 2;
	    need_mood = 0;
	    
	    continue;  // Loop over from start.
	}   

	WATCHDOG_OK;

	/* If both channels are resting then go to deep sleep.	  */
	/* We shouldn't get to here unless all our housekeeping   */
	/* work is done.					  */
	if (ch1resting) {
	    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	    sleep_mode();
	} else {
	    set_sleep_mode(SLEEP_MODE_IDLE);
	    sleep_mode();
	}   
    }		
}

void showbootup (void) {
    // This is a bit of a mystery
    // Wasn't able to find out why the program would
    // break if I tried to something like 
    //	     Song *ch1song = &bootsong;
    const Song *ch1song = bootsongs[0];
    uint16_t ch1noteptr;

    cli();
    ch1restcount = 2;
    ch1resting = 1; 
    sei();

    for (ch1noteptr=0; ch1noteptr < ch1song->notecount; ch1noteptr++) {
	ch1nextnote = pgm_read_byte(&ch1song->notes[ch1noteptr++]);
	ch1noteready = 1;

	while (ch1noteready == 1) {
	    WATCHDOG_OK;
	    set_sleep_mode(SLEEP_MODE_IDLE);
	    sleep_mode();
	}
    }
    
    cli();
    ch1restcount = 1;	
    sei();

    while (ch1playing) {
	WATCHDOG_OK;
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_mode();
    }
    
}

uint8_t randwaitval (void) {
#ifdef PHOTOSHOOT
    return mood + randbits(3) + randbits(3)  + randbits(3); 
#else
    return mood + randbits(4) + randbits(4) + randbits(3) + 
	   randbits(2) + randbits(2) - randbits(3); 
#endif
}

uint8_t randscaleval (void) {

#ifdef BRIGHTER
    return (randbits(6) + randbits(5) + randbits(5) + randbits(4) + 
	    randbits(3) + 54);
#elif defined(PHOTOSHOOT)
    // for photoshoot
    return (randbits(6) + randbits(5) + randbits(4) + randbits(3) + 
	    randbits(2) + 34);
#else 
    /* Returns bell-like distributed random value between %23 - 69% centered
	around 46% */
    return (randbits(6) + randbits(5) + randbits(4) + randbits(3) + 
	    randbits(2) + 58);
#endif
}



/* randbits -- implimentation of linear feedback shift register. */
uint8_t randbits (uint8_t bits) {
    static uint16_t lfsr = RANDOM_SEED;
    
    if (lfsr == 0) FREAKOUT;

    if ((lfsr&0x8000)==0)
	lfsr=lfsr<<1;
    else {
	lfsr=lfsr<<1;
	lfsr = lfsr ^ 0x1D87;
    }
    
    // Mask out the bits we want.
    return ((lfsr & 0xFF) & pgm_read_byte(&bitmasks[bits]));	
}

const Song *pickasong (void) {
    // when the number of songs to choose from is a power of two, the
    // optimizer makes this quite fast.  Otherwise it can take hundreds of cycles to
    // do the multiplication.
    return songs[ (randbits(8) * (sizeof(songs)/sizeof(Song *))) >> 8 ];
}

